#!/usr/local/bin/perl

#------------------------------------------------------------------------
#
#  IBM Software
#  Licensed Material - Property of IBM
#  (C) Copyright International Business Machines Corp. 2007, 2017
#  All Rights Reserved.
#
#  U.S. Government Users Restricted Rights - Use, duplication or
#  disclosure restricted by GSA ADP Schedule Contract with
#  IBM Corp.
#
#------------------------------------------------------------------------

use File::Basename;   # fileparse

$force = 0;

if ( $#ARGV == 1 )
{
    if ( $ARGV[0] eq "-f" )
    {
        $force = 1;
        $fullexecname = $ARGV[1];
    } else {
        print "Usage (args=$#ARGV): parse_dwarf {-f} <elf-image>\n";
        exit 1;
    }
} 
else 
{
    if ( $#ARGV != 0 )
    {
        print "Usage (args=$#ARGV): parse_dwarf {-f} <elf-image>\n";
        exit 1;
    }

    $fullexecname = $ARGV[0];
}

sub get_new_state {
    my $line = shift;

    if ( $line =~ /(DW_TAG_structure_type)/ ) {
        return TAG;
    } elsif ( $line =~ /DW_AT_name/ ) {
        return NAME;
    } elsif ( $line =~ /DW_TAG_member/ ) {
        return STRUCT;
    } elsif ( $line =~ /DW_AT_declaration/ ) {
        return STRUCT_DECL;
    } elsif ( $line =~ /DW_AT_data_member_location/ ) {
        return MEMBER;
    }
}
sub get_level {
    my $line = shift;
    my $oldlevel = shift;

    if ( $line =~ /Abbrev Number/ ) {
        $line =~ s/\>.*//;
        $line =~ s/<//;
        return $line;
    }
    return $oldlevel
}

sub load_dwarf { 

    $array_name=shift;
    $image=shift;
    $outfile=shift;

    open( DWARF, "readelf -wip `readlink -f $image` |" ) or die "Can't open input dwarf pipe\n" ;

    $state=startup;
    $level=1;

    while ( <DWARF> ) {

        $line=$_;
        $line=~s/\n//;

        $level=get_level($line,$level);
        $nextstate=get_new_state($line);

        if ( $state eq "STRUCT" ) {
                #
                # OK.  We found the start of a structure member
                #
                if ( $nextstate eq "NAME" ) {
                    #
                    #   In case we found a new struct
                    #
                    $member_name=$line;
                    $member_name=~s/.*://;
                    $member_name=~s/^\s*//;
                    $member_name=~s/\s*$//;
                } elsif ( $nextstate eq "MEMBER" ) {
                    $member_offset=$line;
                    $member_offset=~s/.*://;
                    $member_offset=~s/\)$//;
                    #puts "struct member name $struct_name (member $member_name) offset $member_offset"
                    #print $outfile "set vmlinux_structs($struct_name.$member_name) $member_offset\n";
                    $key = "$struct_name{$level}.$member_name";
                    $struct_info{$key} = $member_offset;
                } elsif ( $nextstate eq "TAG" ) {
                    $state=$nextstate;
                    #puts "yet another struct found"
                } else {
                    #puts "state: STRUCT - new=$nextstate"
                }
        }

        elsif ( $state eq "TAG" ) {
                #
                # OK.  We found a start of a structure
                #
                if ( $nextstate eq "NAME" ) {
                    $new_struct_name=$line;
                    $new_struct_name=~s/.*://;
                    $new_struct_name=~s/^\s*//;
                    $new_struct_name=~s/\s*$//;
                    #puts "struct name $new_struct_name found"
                } elsif ( $nextstate eq "STRUCT_DECL" ) {
                    #
                    # If we found a member name without finding a DW_AT_declaration then we have a new structure
                    # Just a declaration of a struct pointer within a struct...ignore it.
                    #
                    $new_struct_name="";
                } elsif ( ( $nextstate eq "STRUCT" ) || ( $nextstate eq "TAG" ) ) {
                    #puts "struct member found"
                    $state=$nextstate;

                    if ( $new_struct_name ne "" ) {
                        $struct_name{$level} = $new_struct_name;
                    }
                } else {
                    #puts "state: TAG - new=$nextstate"
                }
        } elsif ( $state eq "startup" ) {
                #
                # When we begin, we just look for a structure
                #
                if ( $nextstate eq "TAG" ) {
                    #puts "new tag structure found!!!"
                    $state=$nextstate;
                } else {
                    #puts "state: startup - new=$nextstate"
                }
         } 
    }

    foreach $key (sort keys %struct_info) {
        #print $outfile "set vmlinux_structs($struct_name.$member_name) $member_offset\n";
        $key_str = $key;
        $key_str =~ s/ /_/g;
        print $outfile "set ${array_name}($key_str) $struct_info{$key}\n";
    }
    close ( DWARF );
    close ( $outfile );
}

sub find_dwarf_info {

    $fullexecname = shift;
    ($name,$path,$suffix) = fileparse($fullexecname,@suffixlist);

    @structinfo_files = ("${fullexecname}.structinfo.tcl","${name}.structinfo.tcl");

    # First, let's see if there is already a structinfo file
    if ( $force == 0 ) {
        foreach $checkfile (@structinfo_files) {
            if ( -r $checkfile ) {
                return ( $checkfile );
            } else {
                #print "$checkfile is not readable or does not exist...\n";
            }
        }
    }

    # Bummer... we didn't find one that exists.   Let's find where we can put one we generate
    $newfilename="";

    foreach $checkfile (@structinfo_files) {
        if ( $newfilename eq "" )
        {
            if ( ! (open(STRUCTINFO, ">$checkfile")) ) {
                $newfilename = "";
                #print "$checkfile is not writable...\n";
            } else {
                $newfilename = $checkfile
            }
        }
    }

    # If we found a file to write to, generate the file
    if ( $newfilename ne "" ) {
        #print "Generating structinfo data into file $newfilename...\n";
        load_dwarf( vmlinux_structs, $fullexecname, STRUCTINFO );
        #print "Done Generating structinfo data.\n";
        return "$newfilename";
    } else {
        print "FAILED: Failed to find an appropriate output file destination\n";
        return "";
    }

}

$dwarf_info_file = find_dwarf_info ( $fullexecname );

print "$dwarf_info_file\n";
