#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2008, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  epapr.tcl
#
# This file will generate a flat device tree blob that represents
# the Openfirmware devtree tha Mambo created
#
# see the ePAPR or by the Device Tree Compiler:
#   http://www.jdl.com/software/)
# for the format
#

namespace eval epapr {
    variable mp
    variable sp 0
    # we use an array because I'm lazy
    array set strtab {}

proc mch { sim ch } {
    variable mp

    $sim memory set $mp 1 [scan $ch %c]

    incr mp
}

proc mb { sim byte } {
    variable mp

    $sim memory set $mp 1 $byte

    incr mp
}

proc mw { sim word } {
    variable mp

    $sim memory set $mp 4 $word
    set mp [expr $mp+4]
}

proc md { sim dword } {
    variable mp

    $sim memory set $mp 8 $dword
    set mp [expr $mp+8]
}

proc mpad { sim s } {
    variable mp

    switch $s 4 { set m 0x3 } 8 { set m 0x7 }

    # pad to word
    while { [expr $mp & $m] } {
	$sim memory set $mp 1 0
	incr mp
    }
}

proc strtab_add { str } {
    variable sp
    variable strtab

    set ret $sp

    foreach c [split $str ""] {
	set strtab($sp) $c
	incr sp
    }
    set strtab($sp) ""
    incr sp

    return $ret
}

proc strtab_dump { sim } {
    variable strtab
 
    for { set i 0 } { $i < [array size strtab ] } { incr i } {
	mch $sim $strtab($i)
    }
}

proc end_node { sim } { mw $sim 2 }
proc nop { sim } { mw $sim 4 }
proc end_dt { sim } { mw $sim 9 }

proc phandle { sim dev_node } {
    # write phandle
    mw $sim 3
    mw $sim 4
    mw $sim [ strtab_add "phandle" ]
    mw $sim $dev_node
    # write linux,phandle
    mw $sim 3
    mw $sim 4
    mw $sim [ strtab_add "linux,phandle" ]
    mw $sim $dev_node
}

proc mstrcpy { sim str } {
    variable mp

    # copy string into memory
    foreach c [split $str ""] {
	mch $sim $c
    }
    $sim memory set $mp 1 0
    incr mp

    mpad $sim 4
}

proc begin_node { sim name } {
    mw $sim 1
    mstrcpy $sim $name
}

proc get_driver_name { path } {
    set i [ string last "/" $path ]
    if { 0 == $i  && [ string length $path ] == 1 }  then {
        set name "/"
    } else {
        incr i
        set name [ string range $path $i end ]
    }
    return $name;
}

# We "fixup" properties, some we drop (by returning false) but the
# most difficult is we need to identify properties with <phandle>s and
# fix the up. Unfortunately, there is no automagic way to figure these
# out.

proc fix_prop { sim dev_name prop } {
    switch $prop {
	name { return false }
	device_type {
	    switch -regexp $dev_name {
		(^serial@[0-9a-f]+$) { return true }
		(^memory@[0-9a-f]+$) { return true }
		(^PowerPC@[0-9a-f]+$) { return true }
		(^interrupt-controller$) - (^interrupt-controller@[0-9a-f]+$) {
		    return true
		}
		default {return false }
	    }
	}
    }
    switch -regexp $dev_name {
	(^memory@[0-9a-f]+$) {
	    switch -regexp $prop {
		(^available$) { return false }
	    }
	}
	(^chosen$) {
	    switch -regexp $prop {
		(^stdout$) {
		    uplevel {
			set fp_chosen [$sim of find_device "/chosen"]
			set fp_std [$sim of getprop $fp_chosen "stdout"]
			set fp_pstd [$sim of package_to_path $fp_std]
			set prop_type "str"
			set prop_value $fp_pstd
			set prop_len [ expr [ string length $prop_value ] + 1 ]
			set prop "linux,stdout-path"
		    }
		    return true
		}
		default { return true }
	    }
	}
	(^cpus$) {
	    switch $prop {
		smp-enabled {return false}
	    }
	}
	(^PowerPC@[0-9a-f]+$) {
	    switch -regexp $prop {
		(^next-level-cache$) {
		    uplevel {
			set prop_value [format %x $prop_value]
			set prop_type "phandle"
		    }
		    return true
		}
	    }
	}
    }
    return true
}

proc do_prop { sim dev_name prop prop_value prop_type prop_len } {

    if { ! [fix_prop $sim $dev_name $prop ] } { return }

    # write the property node id
    mw $sim 3

    # write the size
    mw $sim $prop_len
    # write it to the string tab
    mw $sim [ strtab_add $prop ]

    switch $prop_type {
	phandle {
	    mw $sim $prop_value
	}
	int {
	    mw $sim $prop_value
	}
	str {
	    mstrcpy $sim $prop_value
	}
	byte_array {
	    foreach {hi lo} [split $prop_value {} ] {
		mb $sim "0x$hi$lo"
	    }
	    mpad $sim 4
	}
	array {
	    for { set i 0 } { $i < [ llength $prop_value ] } { incr i } {
		mw $sim [ lindex $prop_value $i ]
	    }
	}
	array64 {
	    for { set i 0 } { $i < [ llength $prop_value ] } { incr i } {
		md $sim [ lindex $prop_value $i ]
	    }
	}
	empty {
	    # nothing
	}
	none {
	    # nothing
	}
	enum {
	    error "cannot handle enum properties"
	}
	default {
	    error "Unknown propert type: $prop_type"
	}
    }
}

proc do_props { sim dev_name dev_node } {
    set prop [ $sim mcm 0 of nextprop $dev_node ]
    set count 0

    phandle $sim $dev_node

    while { $prop != "@null" } {
        set prop_value [ $sim mcm 0 of getprop $dev_node $prop ]
        set prop_type [ $sim mcm 0 of getproptype $dev_node $prop ]
        set prop_len [ $sim mcm 0 of getproplen $dev_node $prop ]

        do_prop $sim $dev_name $prop $prop_value $prop_type $prop_len
        set prop [ $sim mcm 0 of nextprop $dev_node $prop ]
        incr count
        if { $count > 500 } {
	    puts "Endless array: $dev_name"
	    exit 1
        }
    }
}

#
# We get rid of useless nodes here
#
proc prune_node { path } {
    switch $path {
	/openprom -
	/options { return true }
    }
    return false
}

proc do_device { sim dev_node } {
    set name [ $sim mcm 0 of package_to_path $dev_node ]

    # We reverted back to using PowerPC in order to enable
    # ppc64_cpu command to work properly.
    # TODO: Figure out the reason why the code commented below was
    # added in the first place
    #
    # Fix cpu nodes to be "cpu@x" instead of "PowerPC@x"
    # Also kill the leading 0's in x
    # if { [regexp {^/cpus/PowerPC@([0-9a-fA-f]+$)} $name -> cpuno] } {
    #     set name "/cpus/cpu@[format %x 0x$cpuno]"
    # }

    set dev_name [get_driver_name $name]

    if { $dev_name == "/" } {
	set dev_name ""
    }

    if {[ prune_node $name ]} { return }

    begin_node $sim $dev_name

    do_props $sim $dev_name $dev_node

    set child [$sim mcm 0 of child $dev_node ]
    while { $child != 0 } {
        do_device $sim $child
        set child [ $sim mcm 0 of peer $child ]
    }
    end_node $sim
}

proc of2dtb { sim mem } {
    variable mp
    variable sp

    set mp $mem
    set sp 0

    set dev "/"
    set dev_node [ $sim mcm 0 of find_device $dev ]
    if { $dev_node <= 0 } {
        puts "device $dev not found"
        return 0
    }

    #
    # Do the header
    #
    # 0x00
    mw $sim 0xd00dfeed;			# magic
    mw $sim 0;				# totalsize (later)
    mw $sim 0xfeedface;			# off_dt_struct (later)
    mw $sim 0xfeedface;			# off_dt_strings (later)
    # 0x10
    mw $sim 0xfeedface;			# off_mem_rsvmamp (later)
    mw $sim 17;				# version
    mw $sim 16;				# last_comp_version
    mw $sim 0;				# boot_cpuid_phys
    # 0x20
    mw $sim 0xfeedface;			# size_dt_strings (later)
    mw $sim 0xfeedface;			# size_dt_struct  (later)
    nop $sim;				# pad
    nop $sim;				# pad
    # 0x30
    #
    # DT_STRUCTS start here
    #
    set dt_start $mp
    # off_dt_struct
    $sim memory set [expr $mem+0x08] 4 [expr $mp-$mem]
    if { [catch {do_device $sim $dev_node} errs] } {
        puts "do_device got error $errs\n"
    }
    end_dt $sim

    # size_dt_struct
    $sim memory set [expr $mem+0x24] 4 [expr $mp-$dt_start]
    nop $sim;			# pad
    nop $sim;			# pad

    #
    # strtab
    #
    set str_start $mp
    # off_dt_strings
    $sim memory set [expr $mem+0x0c] 4 [expr $mp-$mem]
    strtab_dump $sim
    # size_dt_strings
    $sim memory set [expr $mem+0x20] 4 [expr $mp-$str_start]

    #
    # Reservation map
    #
    mpad $sim 8
    $sim memory set [expr $mem+0x10] 4 [expr $mp-$mem]
    md $sim 0;			# rsvmap empty end entry
    md $sim 0;			# rsvmap empty end entry

    set size [expr $mp-$mem]
    # total size
    $sim memory set [expr $mem+0x04] 4 $size

    return $size
}

proc rsvmap_add { {sim mysim} mem addr size } {
    variable mp

    set rsv_off [ $sim memory display [expr $mem+0x10] 4 ]
    set rsv [expr $mem+$rsv_off]

    # Find last resv map
    while { true } {
	set d0 [ $sim memory display $rsv 8 ]
	set d1 [ $sim memory display [expr $rsv+8] 8 ]
	if { [expr $d0 | $d1] == 0 ] } {
	    break
	}
	set rsv [expr $rsv+16]
    }
    set mp $rsv
    md $sim $addr
    md $sim $size
    md $sim 0;			# rsvmap empty end entry
    md $sim 0;			# rsvmap empty end entry

    # update header
    set size [expr $mp-$mem]
    # total size
    $sim memory set [expr $mem+0x04] 4 $size

    return $size
}

proc epapr { { sim mysim} mem ima_size args } {

    set size [ of2dtb $sim $mem ]

    $sim cpu 0 set gpr 1 0
    $sim cpu 0 set gpr 2 0
    $sim cpu 0 set gpr 3 $mem
    $sim cpu 0 set gpr 4 0
    $sim cpu 0 set gpr 5 0
    # ePAPR booke magic number
    $sim cpu 0 set gpr 6 0x45504150
    $sim cpu 0 set gpr 7 $ima_size
    $sim cpu 0 set gpr 8 0
    $sim cpu 0 set gpr 9 0

    return $size
}

proc dump { { sim mysim } file { dtb -1 } } {
    if { $dtb == -1 } { set dtb [$sim cpu 0 display gpr 3] }
    set size [ $sim memory display [expr $dtb + 4] 4]

    $sim memory fwrite $dtb $size $file
}

}
