#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003, 2017
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  openfirmware_utils.tcl
#
#
#------------------------------------------------------------------------

namespace eval of {

proc show_prop { prop prop_value pad chid } {
    if {[ llength $prop_value] == 1} {
        puts $chid [ format "%s%-30s %s" $pad $prop [ lindex $prop_value 0 ] ]
    } else {
        set prop_line [ format "%s%-30s < %s" $pad $prop [ lindex $prop_value 0 ] ]
        for { set i 1 } { $i < [ llength $prop_value ] } { incr i } {
            set prop_line "$prop_line [ format "%s" [ lindex $prop_value $i ] ]"
        }
        set prop_line "$prop_line >"
        puts $chid $prop_line
    }
}

proc get_driver_name { path } {
    set i [ string last "/" $path ]
    if { 0 == $i  && [ string length $path ] == 1 }  then {
        set name "root"
    } else {
        incr i
        set name [ string range $path $i end ]
    }
    return $name;
}


proc show_props { sim dev_node indent chid } {
    set prop [ $sim mcm 0 of nextprop $dev_node ]
    set count 0
    set pad [string repeat " " [ expr $indent + 4 ] ]
    while { $prop != "@null" } {
        set prop_value [ $sim mcm 0 of getprop $dev_node $prop ]
        show_prop $prop $prop_value $pad $chid
        set prop [ $sim mcm 0 of nextprop $dev_node $prop ]
        incr count
        if { $count > 500 } {
            exit
        }
    }
}

proc show_device { sim dev_node indent chid } {
    set name [ $sim mcm 0 of package_to_path $dev_node ]
    puts $chid "[string repeat " " $indent][get_driver_name $name] ([format "%x" $dev_node])"
    show_props $sim $dev_node 4 $chid
    set child [$sim mcm 0 of child $dev_node ]
    while { $child != 0 } {
        show_device $sim $child [ expr $indent + 4 ] $chid
        set child [ $sim mcm 0 of peer $child ]
    }
    puts $chid " "
}

proc show_tree { { dev "/" } { fhandle stdout } { sim mysim }} {
    set dev_node [ $sim mcm 0 of find_device $dev ]
    if { $dev_node <= 0 } {
        puts "device $dev not found"
        return
    }
    if { [catch {show_device $sim $dev_node 0  $fhandle} errs] } {
        puts "show_device got error $errs\n"
    }
}

proc list_devices {{ dev "/" } { sim mysim }} {
    set dev_node [ $sim mcm 0 of find_device $dev ]
    set child [$sim mcm 0 of child $dev_node ]
    set path [ $sim mcm 0 of package_to_path $dev_node ]

    lappend out $path
    while { $child != 0 } {
	set path [ $sim mcm 0 of package_to_path $child ]
        lappend out [list_devices $path $sim]
        set child [ $sim mcm 0 of peer $child ]
    }
    return [join $out "\n"]
}

proc save_tree { { dev "/" } { fname dev_tree.txt } { sim mysim } } {

    set fhandle [ open $fname "w" ]
    if { $fhandle != -1 } {
        show_tree $dev $fhandle $sim
        close $fhandle
    }
}

proc add_rtas_function { dev_node name { sim mysim } { mcm 0 } } {
    set token [ $sim mcm $mcm of rtasinfo $name ]
    if { $token == -1 } {
        error "rtas token not found, function= $name"
    }
    $sim mcm $mcm of addprop  $dev_node "int"  $name $token
}


proc add_array_prop { dev_node name value { sim mysim } } {
    # note we need to pass the name of the variable, not the value
    $sim mcm 0 of addprop  $dev_node "array"  $name "value"
}

proc add_machine_info_props { dev_node { sim mysim } { mcm 0 } } {

    $sim mcm $mcm of machineinfo
    set props [ array get machine_info ]
    for { set i 0 } { $i < [ llength $props ] } { set i [ expr $i+2 ] } {
        $sim mcm $mcm of addprop $dev_node "int" [ lindex $props $i ] [ lindex $props [ expr $i + 1 ] ]
    }
}

proc add_sim_node { { sim mysim } { mcm 0 } } {

    set root_node [ $sim mcm $mcm of peer 0 ]

    regexp {(SystemSim|Mambo) Version (\d+)\.(\d+)[\.-](\d+)} [version list] match v0 v1 v2 v3
    set version [expr ($v1<<24)|($v2<<16)|($v3<<8)|2]

    # Systemsim node for bogus devices
    set sim_node [$sim mcm $mcm of addchild $root_node "systemsim" ""]
    $sim mcm $mcm of addprop $sim_node "string"  "name"  "systemsim"
    $sim mcm $mcm of addprop $sim_node "int"  "\#address-cells"  1
    $sim mcm $mcm of addprop $sim_node "int"  "version"  $version
}

# Use of the Mambo node is deprecated but the node is retained for compatability with
# previous versions of the bogus devices.
proc add_mambo_node { { sim mysim } { mcm 0 } } {

    set root_node [ $sim mcm $mcm of peer 0 ]

    regexp {(SystemSim|Mambo) Version (\d+)\.(\d+)[\.-](\d+)} [version list] match v0 v1 v2 v3
    set version [expr ($v1<<24)|($v2<<16)|($v3<<8)|2]

    # Mambo node for bogus devices
    set sim_node [$sim mcm $mcm of addchild $root_node "mambo" ""]
    $sim mcm $mcm of addprop $sim_node "string"  "name"  "mambo"
    $sim mcm $mcm of addprop $sim_node "int"  "\#address-cells"  1
    $sim mcm $mcm of addprop $sim_node "int"  "version"  $version
}

proc set_bootargs { argstring { sim mysim } { mcm 0 } } {
    set chosen_node [$sim mcm $mcm of find_device "/chosen"]
    if {$chosen_node >= 0} {
        if {[string length [$sim mcm $mcm of getprop $chosen_node "bootargs"]] == 0} {
            $sim mcm $mcm of addprop $chosen_node "string"  "bootargs"  $argstring
        } else {
            $sim mcm $mcm of setprop $chosen_node "bootargs"  $argstring
        }
    }
}

proc mmap_info { map_name { sim mysim } { mcm 0 } } {
    return [$sim mcm $mcm of mmapinfo $map_name ]
}

proc unit_address { mapinfo } {
    set addr [ format "%x" [ lindex $mapinfo 1 ] ]
    set hi [ lindex $mapinfo 0 ]
    if { 0 != $hi } {
        set addr [ format "%x" $hi ]$addr
    }
}

proc change_open_firmware_block_size { { sim mysim } { size 128 } } {
  for {set i 0} {$i < [$sim display cpus]} {incr i} {
    set x [format "%x" $i]
    $sim of setprop [$sim of find_device /cpus/PowerPC@$x] d-cache-block-size $size
    $sim of setprop [$sim of find_device /cpus/PowerPC@$x] d-cache-line-size $size
    $sim of setprop [$sim of find_device /cpus/PowerPC@$x] i-cache-block-size $size
    $sim of setprop [$sim of find_device /cpus/PowerPC@$x] i-cache-line-size $size
  }
}

# Encode a multi-string compatible value as a byte array for OF
proc encode_compat { compat_list } {
    set barray [list]
    foreach str $compat_list {

        for { set i 0 } { $i < [string length $str]} {incr i } {
            set char [string index $str $i]
            scan $char "%c" ord
            lappend barray [format "%2x" $ord]
        }
        lappend barray "00"
    }

    return $barray
}

}
