#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2005, 2017
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#   linux.tcl
#------------------------------------------------------------------------

namespace eval Linux_Utils {
    variable version            "1.0"
    variable vmlinux_log_line   ""
}

set process_track_verbose 0

package provide Linux_Utils $Linux_Utils::version

# ----------------------------------------------------------------------
# Utility functions to load linux and sysroot disk
# ----------------------------------------------------------------------

proc load_vmlinux_and_rootdisk { { machine p7 } { sim mysim } { linux_dir linux } } {
    global env

    set img_dir $env(IMAGES_DIR)/$machine/$linux_dir

    # Find the OS image
    if {[file exists ./vmlinux]} {
        set vmlinux_file ./vmlinux
    }  elseif {[file exists $img_dir/vmlinux.img]} {
        set vmlinux_file $img_dir/vmlinux.img
    } else {
        puts "could not find linux image file: $img_dir/vmlinux.img"
        exit
    }

    # Find the rootdisk to use
    if {[file exists ./rootdisk]} {
        set rootdisk_file ./rootdisk
    }  elseif {[file exists $img_dir/rootdisk.img]} {
        set rootdisk_file $img_dir/rootdisk.img
    } else {
        puts "could not find rootdisk file"
        exit
    }

    # Set boot parameters if desired
    of::set_bootargs "root=/dev/mambobd0 rw console=hvc0"

    # Load the OS
    $sim load vmlinux $vmlinux_file 0x0

    # Setup the bogus disk that Linux will use
    if {[catch {$sim bogus disk init 0 $rootdisk_file newcow}]} {
        puts "bogus disk init for $rootdisk_file failed"
        exit
    }
}

# ----------------------------------------------------------------------
# Utility function to load a linux initrd
# ----------------------------------------------------------------------

proc load_initrd { {machine pegasus } { mem_addr 0x00000000033b6000} { initrd initrd.img } { linux_dir linux } { add_dt_entries true } } {
    global env

    set img_dir $env(IMAGES_DIR)/$machine/$linux_dir

    # Find the initrd to use
    if {[file exists ./$initrd]} {
        set initrd_file ./$initrd
    } elseif {[file exists $img_dir/$initrd]} {
        set initrd_file $img_dir/$initrd
    } else {
        puts "could not load initrd file: $initrd"
        exit
    }

    set initrd_size [file size $initrd_file]

    puts "copying $initrd_size bytes of initrd at $initrd_file to $mem_addr"
    mysim memory fread $mem_addr $initrd_size $initrd_file

    if {[expr $add_dt_entries]} {
        set of_chosen [mysim of find_device /chosen]
        set end_addr [expr $mem_addr + $initrd_size]
        mysim of addprop $of_chosen int "linux,initrd-start" $mem_addr
        mysim of addprop $of_chosen int "linux,initrd-end" $end_addr
    }
}

# ----------------------------------------------------------------------
# advance simulator cycles to align with next epoch.
# epoch value must be a power of 10
# ----------------------------------------------------------------------
proc cycle_align { epoch {sim mysim} {mcm 0} } {
   set curcycle  [$sim display cycles ]
   set epochlen  [expr [string length $epoch ] - 2 ]
   set curoffset [string range $curcycle end-$epochlen end ]      ;# get least signficant digits
   set curoffset [string trimleft $curoffset 0 ]                  ;# get rid of leading 0's
   set delaycycle [expr $epoch - $curoffset ]

   if {[simstop status] == "not-stopped"} {
       $sim mcm $mcm trigger set cycle [expr $curcycle + $delaycycle] { simstop }
   } else {
       $sim cycle $delaycycle
   }
}

# This function will clear all console triggers that invoke a given trigger proc
proc clear_console_trigger { sim mcm proc } {

    # get rid of any console triggers for proc
    set triggers [split [$sim mcm $mcm trigger display console list] "\n"]
    for {set i 0} {$i < [llength $triggers]-1} {incr i} {
        if {[regexp "(\".+\") -> \{$proc" [lindex $triggers $i] match string]} {
            $sim mcm $mcm trigger clear console $string
        }
    }
}

proc Linux_showed_prompt { sim mcm args } {

    # get rid of trigger
    clear_console_trigger $sim $mcm Linux_showed_prompt 

    simstop
}

proc Linux_init_done { sim mcm args } {

    # get rid of trigger
    clear_console_trigger $sim $mcm Linux_init_done

    $sim mcm $mcm trigger set console "#" "Linux_showed_prompt $sim $mcm"
}

proc fast_Linux_init_done { sim mcm args } {

    # get rid of trigger
    clear_console_trigger $sim $mcm fast_Linux_init_done

    $sim mcm $mcm trigger set console "#" "Linux_showed_prompt $sim $mcm"

    $sim mode simple
}

proc wait_for_linux_prompt { { mcm 0 } { sim mysim } } {

    $sim mcm $mcm trigger set console "INIT: Entering runlevel" "Linux_init_done $sim $mcm"
    $sim mcm $mcm trigger set console "Starting login process" "Linux_init_done $sim $mcm"
    $sim go
}

proc wait_for_linux_pound_sign { { mcm 0 } { sim mysim } } {
    wait_for_linux_prompt $mcm $sim
}

proc wait_for_just_linux_pound_sign { { mcm 0 } { sim mysim } } {

    $sim mcm $mcm trigger set console "#" "Linux_showed_prompt $sim $mcm"
    $sim go
}

proc fastest_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
    $sim mcm $mcm trigger set console "INIT: Entering runlevel" "fast_Linux_init_done $sim $mcm"
    $sim mcm $mcm trigger set console "Starting login process" "fast_Linux_init_done $sim $mcm"
    $sim mode fastest
    $sim go
}

proc fast_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
    $sim mcm $mcm trigger set console "INIT: Entering runlevel" "fast_Linux_init_done $sim $mcm"
    $sim mcm $mcm trigger set console "Starting login process" "fast_Linux_init_done $sim $mcm"
    $sim mode fast
    $sim go
}

proc turbo_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
    $sim mcm $mcm trigger set console "INIT: Entering runlevel" "fast_Linux_init_done $sim $mcm"
    $sim mcm $mcm trigger set console "Starting login process" "fast_Linux_init_done $sim $mcm"
    $sim mode turbo
    $sim go
}

proc kernel_panic_called {sim args} {

    simstop
    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }
    set r3 [$sim cpu $procid display gpr 3]
    set cycle [$sim display cycles]

    # Get string pointed to by R3
    if { [catch "set pa [$sim cpu $procid dtranslate $r3]" errs] == 0 } {
        if { [catch "set panic_msg \"[$sim cpu $procid memory display $pa STRING]\"" errs] != 0 } {
            puts "errors : $errs"
        }
    }
    set pc [$sim cpu $procid display spr pc]
    set ra [get_imem_ra $sim $pc $procid]
    puts "\nWARNING: CPU=$procidd PC=$pc: (real-addr:$ra) VMLINUX PANIC CALLED with MSG: ($panic_msg)\n"
    #mcm_display_gprs $sim $mcm $cpu
}

set vmlinux_filename "vmlinux-not-loaded"

###############################################################
#
# Set a trigger at the START of a specified vmlinux function
#
proc vmlinux_set_trigger { function_name { tcl_proc "break_point mysim $function_name" } { sim mysim } } {
    global vmlinux_filename 

    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
            if { [info exists func_info(FUNC_START)] } {
                set pidtracker_entry $func_info(FUNC_START)
                $sim trigger set pc $pidtracker_entry "$tcl_proc"
            #} else {
            #    puts "no func start"
            }
        } else {
            puts "vmlinux_set_trigger failed with $errs"
        }
    } else {
        puts "vmlinux_set_trigger object call failed with $errs"
    }
}

###############################################################
#
# Set a trigger on all RETURNS (blr) from a specified vmlinux function
#
proc vmlinux_set_trigger_on_return { function_name { tcl_proc "break_point mysim $function_name" } { sim mysim } } {
    global vmlinux_filename 

    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
            if { [info exists func_info(FUNC_START)] } {
                set inst_data [object inst_data $vmlinux_filename $func_info(FUNC_START) $func_info(FUNC_END)]
                array set ea_inst $inst_data
                foreach ea [array names ea_inst] {
                    if { $ea_inst($ea) == 0x4e800020 } {
                        $sim trigger set pc $ea "$tcl_proc"
                    }
                }
            }
        } else {
            puts "vmlinux_set_trigger_on_return failed with $errs"
        }
    } else {
        puts "vmlinux_set_trigger_on_return object call failed with $errs"
    }
}

proc do_nothing {args} {
    #puts "do_nothing called"
}

proc profile {routine args} {
    global profile_buckets

    set start [clock clicks]
    eval "$routine $args"
    set end [clock clicks]

    if { [info exists profile_buckets($routine)] } {
        incr profile_buckets($routine) [expr $end - $start]
    } else {
        set profile_buckets($routine) 0
    }
}

proc dump_profile { } {
    global profile_buckets

    foreach routine [array names profile_buckets] {
        puts "$routine: $profile_buckets($routine) clicks"
    }
}

proc vmlinux_loaded {sim filename address size} {
    
    global vmlinux_filename 

    if { [$sim config query pid_tracking] == off } {
       return
    }


    set vmlinux_filename [file tail $filename]
    vmlinux_get_struct_info $filename

    set do_nothing 0
    set profile "profile "
    #set profile ""
    #
    #   We want to set a breakpoint on .panic
    #   Extract information on symbol from object file and set a breakpoint
    #
    set function_name .panic
    set tcl_proc "$profile kernel_panic_called $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking
    set function_name .__switch_to
    set tcl_proc "$profile pidtracker_switch $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking
    set function_name .do_execve
    set tcl_proc "$profile pidtracker_execve $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim
    set function_name .compat_do_execve
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking
    set function_name .do_mmap_pgoff
    set tcl_proc "$profile pidtracker_do_mmap_pgoff $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking
    # .do_exit is called by .do_exit
    set function_name .do_exit
    set tcl_proc "$profile pidtracker_do_exit $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking
    set function_name .copy_process
    set tcl_proc "$profile pidtracker_copy_process $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger_on_return $function_name $tcl_proc $sim

    # Setup up the PID tracking (Kernel Threads)
    set function_name .kthread_create
    set tcl_proc "$profile pidtracker_kthread_create $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger_on_return $function_name $tcl_proc $sim

    # Setup up the PID tracking (Kernel Threads)
    set function_name .kthreadd
    set tcl_proc "$profile pidtracker_kthreadd $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking 
    set function_name .setup_system
    set tcl_proc "$profile pidtracker_setup_system $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    # Setup up the PID tracking (Kernel Threads)
    set function_name .kthread_stop
    set tcl_proc "$profile pidtracker_kthread_stop $sim"
    if { $do_nothing != 0 } { set tcl_proc "do_nothing $sim" }
    vmlinux_set_trigger $function_name $tcl_proc $sim

    $sim processtree addfile 0 $vmlinux_filename 0 $size 0 0 0
    $sim processtree addfile 0 $vmlinux_filename 0xC000000000000000 $size 0 0 0
}


#
# 32-bit value to ascii
#
proc vmlinux_ascii32 { value } {
	set s [format "%c%c%c%c" [expr ( ($value >> 24) & 0xff )] \
	                         [expr ( ($value >> 16) & 0xff )] \
                             [expr ( ($value >> 8) & 0xff )] \
	                         [expr ( $value & 0xff )]]

    while {! [string is graph -failindex bad $s]} {
        set s [string replace $s $bad $bad "."]
    }
	return $s
}

#
# Display the accumulated vmlinux log line ($Linux_Utils::vmlinux_log_line)
#
proc vmlinux_log_dump {args} {
    if {[string length $Linux_Utils::vmlinux_log_line] != 0} {
        puts "LOG: '$Linux_Utils::vmlinux_log_line'"
        set Linux_Utils::vmlinux_log_line ""
    }
}

#
# Capture vmlinux kernel log calls in $Linux_Utils::vmlinux_log_line
#
proc vmlinux_log_capture {{procid 0:0:0} {sim mysim}} {
    set gpr [$sim cpu $procid display gpr 3]
    set char [string range $gpr [expr [string length $gpr] - 2] end]
    if {[string equal $char "0A"]} {
        vmlinux_log_dump
    } else {
        set letter [string index [vmlinux_ascii32 $gpr] end]
        set Linux_Utils::vmlinux_log_line "${Linux_Utils::vmlinux_log_line}${letter}"
    } 
}

#
# Print kernel calls set up via vmlinux_entries (below). Treats ".emit_log_char" specially
# to show log strings
#
proc vmlinux_call_trigger {sim args} {
    set call [lindex $args 0]
    array set p [lindex $args 1]
    if {[string equal $call ".emit_log_char"]} {
        vmlinux_log_capture $p(mcm):$p(cpu):$p(thread) $sim
    } else {
        if {[string equal $call ".__delay"]} {
            vmlinux_log_dump
        }
        puts -nonewline "KERNEL\[$p(mcm):$p(cpu):$p(thread)\]: $call, $p(address) {"
        for {set i 3} {$i < 6} {incr i} {
            set v [$sim cpu $p(mcm):$p(cpu):$p(thread) display gpr $i]
            set ascii [vmlinux_ascii32 $v]
            puts -nonewline "R$i:$v|$ascii,"
        }
        puts "}"
    }
}

#
# Set PC triggers for all entry points in an objdump file given
# Recognizes lines like the following:
#
# c000000000020d00 <.__delay>:
#
# or from an "nm" generated map file
#
# c000000000020d00 T .__delay
# c000000000020d00 t .__delay
#
proc vmlinux_trace {odfile {trig vmlinux_call_trigger} {sim mysim}} {
    if {[catch {open $odfile r} chan]} {
        puts "vmlinux_trace: Can't open '$odfile' for reading"
        return -code error
    }
    puts "vmlinux_trace: Scanning file '$odfile'..."
    
    set trig_count 0
    while {[gets $chan line] >= 0} {
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<([^>]+)>:} $line match addr entry_pt]} {
            $sim trigger set pc 0x$addr [list $trig $sim "$entry_pt"]
            incr trig_count
        } elseif {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+[Tt]\s+(.+)} $line match addr entry_pt]} {
            $sim trigger set pc 0x$addr [list $trig $sim "$entry_pt"]
            incr trig_count
        }
    }
    close $chan
    puts "vmlinux_trace: $trig_count entry points registered from '$odfile'"
}

#
# Print kernel log calls set up via vmlinux_log (below)
#
proc vmlinux_log_trigger {sim args} {
    array set p [lindex $args 0]
    vmlinux_log_capture $p(mcm):$p(cpu):$p(thread) $sim
}

#
# Set PC triggers for the .emit_log_char entry points in an objdump file given
# Recognizes lines like the following:
#
# c000000000020d00 <.emit_log_char>:
# c000000000020d40 <.__delay>:
#
# Associates log dump with the delay entry point, in case we delay without
# seeing an end-of-line
#
proc vmlinux_log {odfile {sim mysim}} {
    if {[catch {open $odfile r} chan]} {
        puts "vmlinux_log: Can't open '$odfile' for reading"
        return -code error
    }
    puts "vmlinux_log: Scanning file '$odfile'..."

    while {[gets $chan line] >= 0} {
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<\.emit_log_char>:} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_trigger $sim]
            puts "vmlinux_log: found .emit_log_char in '$odfile' at address $addr"
        } elseif {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+[Tt]\s+\.emit_log_char} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_trigger $sim]
            puts "vmlinux_log: found .emit_log_char in '$odfile' at address $addr"
        }
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<\.__delay>:} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_dump]
            puts "vmlinux_log: found .__delay in '$odfile' at address $addr"
        } elseif {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+[Tt]\s+\.__delay} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_dump]
            puts "vmlinux_log: found .__delay in '$odfile' at address $addr"
        }
    }
    close $chan
}

####################################################################
#
# Pid Tracker stuff from Gilles.   Modified some by PJB
#
####################################################################

#------------------------------------------------------------------------
# get_dmem_ra
#     Get the real effective address of data
#
#     (Gilles)     
#------------------------------------------------------------------------
proc get_dmem_ra { sim ea procid } {
    set ra [ $sim cpu $procid dtranslate $ea ]
    return $ra
}


#------------------------------------------------------------------------
# read_mem_at
#     Get value at effective address
#
#     (Gilles)     
#------------------------------------------------------------------------
proc read_mem_at { sim procid ea size } {
    if { $size <= 0 || $size > 8 } {
	return "error"
    }
    set val [ $sim cpu $procid mem display $ea $size]
    return $val
}

#------------------------------------------------------------------------
# get_pc
#     Get current pc
#
#     (Gilles)     
#------------------------------------------------------------------------
proc get_pc { sim procid } {
    set pc [ $sim cpu $procid display spr pc ]
    return $pc
}


#------------------------------------------------------------------------
# get_string_at_ea
#     Get string name at ea
#
#     (Gilles)     
#------------------------------------------------------------------------
proc get_string_at_ea {sim ea procid} {
    set ra [$sim cpu $procid dtranslate $ea]
    set curchar [format "%c" [$sim cpu $procid mem display $ra 1]]
    set retstring ""

    set ra [$sim cpu $procid dtranslate $ea]
    set retstring [$sim cpu $procid memory display $ra STRING]
    #while {[string compare $curchar "\0"]} { #0 is ASCII null string terminator
	#    set retstring "$retstring$curchar"
    #    incr ea 1
	#    set ra [$sim cpu $procid thread $tid dtranslate $ea]
    #    set curchar [format "%c" [$sim memory display $ra 1]]
    #}
    return $retstring
}

proc get_lstring_at_ea {sim ea procid tid} {
    set j 0
    set ra [$sim cpu $procid:$tid dtranslate $ea]
    set curchar [format "%c" [$sim mem display $ra 1]]
    set retstring ""
    while { $j < 2000 } { #0 is ASCII null string terminator
	set retstring "$retstring$curchar"
        incr ea 1
	incr j 1
	set ra [$sim cpu $procid:$tid dtranslate $ea]
        set curchar [format "%c" [$sim memory display $ra 1]]
    }
    puts "$retstring"
}
proc pidtracker_switch { sim args } {

    global vmlinux_structs
    global switch_to

    # We need to grab the args passed to __switch_to at the beginning
    # and then use them at the end so we can be more assured we find translations
    # for both task_struct's

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    # By convention, gpr3 holds pointer to task_struct of the old thread
    #set gpr3 $switch_to($sim.$procid.$tid.gpr3)
    set gpr3 [$sim cpu $procid display gpr 3]

    # By convention, gpr4 holds pointer to task_struct of the new thread    
    #set gpr4 $switch_to($sim.$procid.$tid.gpr4)
    set gpr4 [$sim cpu $procid display gpr 4]

    set oldpid [vmlinux_get_current_pid $procid $sim $gpr3]
    set oldname [vmlinux_get_current_name $procid $sim $gpr3]

    set newpid [vmlinux_get_current_pid $procid $sim $gpr4]
    set newname [vmlinux_get_current_name $procid $sim $gpr4]

    if { [catch {
                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "context switch on cpu $procid : ($oldname, pid $oldpid) -> ($newname, pid $newpid)"
                    }

                    $sim cpu $procid osinfo set pid $newpid exec $newname
                    $sim cpu $procid process switch $newpid
            } errs] } { puts "__switch 3 (gpr3=$gpr3 gpr4=$gpr4) had errs : $errs" }

    catch { EventLog::SwitchTo $sim $procid $oldpid $oldname $newpid $newname } errs

    return 0
}

###########################################################################
#
# On Linux/PPC, r13 points to the PACA.   Within the PACA is a pointer to 
# the "current" task_struct.   
#
#   The task_struct is used to get things like PID and COMM (or name)
#
proc vmlinux_get_current { { procid 0:0:0 } { sim mysim } } {
    global vmlinux_structs

    set gpr13 [$sim cpu $procid display gpr 13]

    # Older kernels had a paca_struct.xCurrent field and newer ones have paca_struct.__current field
    if { [info exists vmlinux_structs(paca_struct.__current)] } {
        return [load_struct_ptr $gpr13 paca_struct.__current $sim $procid]
    }
    if { [info exists vmlinux_structs(paca_struct.xCurrent)] } {
        return [load_struct_ptr $gpr13 paca_struct.xCurrent $sim $procid]
    }
    return -1;
}
proc vmlinux_get_parent { task_struct { procid 0:0:0 } { sim mysim } } {
    if { [info exists vmlinux_structs(task_struct.parent)] } {
        return [load_struct_ptr $task_struct task_struct.parent $sim $procid]
    }
    if { [info exists vmlinux_structs(task_struct.real_parent)] } {
        return [load_struct_ptr $task_struct task_struct.real_parent $sim $procid]
    }
    return 0
}

###########################################################################
#
# If a point to current is passed in, we will use that.   Otherwise we will figure it out.
#
proc vmlinux_get_current_pid { { procid 0:0:0 } { sim mysim } { current empty } } {
    global vmlinux_structs
    global task_struct

    if { [string compare $current "empty"] == 0 } {
        set current [vmlinux_get_current $procid $sim]

        if { $current == -1 } {
            return 0
        }
    }

    if { [info exists task_struct($sim.$current.pid)] } {
        return $task_struct($sim.$current.pid)
    }

    if { ! [info exists vmlinux_structs(task_struct.pid)] } {
        return 0
    }

    if { [catch { 
                    set pid_ea [expr $current + $vmlinux_structs(task_struct.pid)]
                    set pid_ra [get_dmem_ra $sim $pid_ea $procid]
                    set pid [read_mem_at $sim $procid $pid_ra 4]

                    set task_struct($sim.$current.pid) $pid
                } errs] } { puts "Errors from vmlinux_get_current_pid: $errs" }

    return [format "%d" $pid]
}

###########################################################################
#
# If a point to current is passed in, we will use that.   Otherwise we will figure it out.
#
proc vmlinux_get_current_ppid { { procid 0:0:0 } { sim mysim } { current empty } } {
    global vmlinux_structs

    if { [string compare $current "empty"] == 0 } {
        set current [vmlinux_get_current $procid $sim]

        if { $current == -1 } {
            return 0;
        }
    }
    if { ! [info exists vmlinux_structs(task_struct.pid)] } {
        return 0 
    }
    set parent [vmlinux_get_parent $current $procid $sim]
    set ppid_ea [expr $parent + $vmlinux_structs(task_struct.pid)]
    set ppid_ra [get_dmem_ra $sim $ppid_ea $procid]
    set ppid [read_mem_at $sim $procid $ppid_ra 4]

    return [format "%d" $ppid]
}

###########################################################################
#
# On Linux/PPC, r13 points to the PACA.   Within the PACA is a pointer to 
# the "current" task_struct.   Within the task_struct is the PID and COMM (or name)
#
# If a point to current is passed in, we will use that.   Otherwise we will figure it out.
#
proc vmlinux_get_current_name { { procid 0:0:0 } { sim mysim } { current empty } } {
    global vmlinux_structs
    global task_struct

    if { [string compare $current "empty"] == 0 } {
        set current [vmlinux_get_current $procid $sim]

        if { $current == -1 } {
            return "unknown"
        }
    }

    if { [info exists task_struct($sim.$current.name)] } {
        return $task_struct($sim.$current.name)
    }

    if { ! [info exists vmlinux_structs(task_struct.comm)] } {
        return "unknown"
    }

    set name_ea [expr $current + $vmlinux_structs(task_struct.comm)]
    set name [get_string_at_ea $sim $name_ea $procid]

    set task_struct($sim.$current.name) $name

    return $name
}

# We can get this from readelf but we will set it by hand for now
proc vmlinux_get_struct_info { vmlinux_filename } {
    global vmlinux_structs
    global env

    #
    #   Execute the parse_dwarf.pl command to return the tcl file that contains the structure info
    #
    set parse_dwarf_file $env(EXEC_DIR)/parse_dwarf.pl

    if { [catch {
                    set tclfile [exec $parse_dwarf_file $vmlinux_filename]
                    source $tclfile
                } errs ] } { puts "vmlinux_get_struct_info got errors: $errs" }

    #set vmlinux_structs(task_struct.parent) 256
    #set vmlinux_structs(task_struct.comm) 648
    #set vmlinux_structs(paca_struct.__current) 432
    #set vmlinux_structs(pid.nr) 4
    #set vmlinux_structs(file.f_path) 16
    #set vmlinux_structs(path.dentry) 8
    #set vmlinux_structs(dentry.d_name) 48
    #set vmlinux_structs(qstr.name) 8
    #set vmlinux_structs(task_struct.pid) 248
    #set vmlinux_structs(task_struct.parent) 256
    #set vmlinux_structs(task_struct.comm) 648
    #set vmlinux_structs(paca_struct.__current) 432
    #set vmlinux_structs(pid.nr) 4
    #set vmlinux_structs(file.f_path) 16
    #set vmlinux_structs(path.dentry) 8
    #set vmlinux_structs(dentry.d_name) 48
    #set vmlinux_structs(qstr.name) 8
}

proc pidtracker_execve { sim args } {

    # Note: the pid/name offset is hardcoded and may changed
    #       with different kernel version. The offsets used
    #       in this code is derived from kernel version 2.6.7    
    global vmlinux_structs

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    set pid [vmlinux_get_current_pid $procid $sim]

    # gpr3 holds pointer to the filename to be exec'd
    set gpr3 [$sim cpu $procid display gpr 3]
    set filename [get_string_at_ea $sim $gpr3 $procid]

    global process_track_verbose
    if { $process_track_verbose > 0 } {
        puts "execve called for pid $pid filename $filename"
    }

    $sim processtree exec $pid $filename
    # Update the EventLog
    if { [catch { EventLog::KernelExec $sim $procid $pid $filename } errs] } { puts "execve has errs $errs" }

    return 0
}

#
# The proc should be called on return from copy_process to r3 points to the new task_struct
#
proc pidtracker_copy_process { sim args } {

    global vmlinux_structs

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    set gpr3 [$sim cpu $procid display gpr 3]

    # If on return it is not an error (ie gpr3 u< -4095) then get the results
    if { $gpr3 < 0xFFFFFFFFFFFFF000 } {
        set pid [vmlinux_get_current_pid $procid $sim $gpr3]
        set ppid [vmlinux_get_current_pid $procid $sim]
        if { [catch "$sim processtree newproc $pid $ppid 0" errs] } { puts "newproc errs are $errs" }
    }
    return 0

    #if { [catch {
    #                set clone_flags [$sim cpu $procid thread $tid display gpr 3]
    #                set stack_start [$sim cpu $procid thread $tid display gpr 4]
    #                set regs [$sim cpu $procid thread $tid display gpr 5]
    #                set stack_size [$sim cpu $procid thread $tid display gpr 6]
    #                set parent_tidptr [$sim cpu $procid thread $tid display gpr 7]
    #                set child_tidptr [$sim cpu $procid thread $tid display gpr 8]
    #                set pid_p [$sim cpu $procid thread $tid display gpr 9]
#
#                    set pid_ea [expr $pid_p+$vmlinux_structs(pid.nr)]
#                    set pid_ra [get_dmem_ra $sim $pid_ea $procid $tid]
#                    set pid [read_mem_at $sim $pid_ra 4]
#
#                    global process_track_verbose
#                    if { $process_track_verbose > 0 } {
#                        puts "copy_process called for clone_flags:$clone_flags stack_start:$stack_start regs:$regs stack_size:$stack_size parent_tidptr:$parent_tidptr child_tidptr:$child_tidptr pid:$pid"
#                    }
#                } errs] } { puts "copy_process had errs : $errs\n" }
#
#
#    if { [catch "$sim processtree newproc $pid [vmlinux_get_current_pid $procid $tid $sim] $clone_flags" errs] } { puts "newproc errs are $errs" }
#
}

#
#   This proc should be called on return from the kthread_create so 
#   r3 has a pointer to the setup task_struct
#
proc pidtracker_kthread_create { sim args } {

    global vmlinux_structs
    global vmlinux_filename

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    if { [catch {
                    set new_task_struct [$sim cpu $procid display gpr 3]
                    set name [vmlinux_get_current_name $procid  $sim $new_task_struct]
                    set pid [vmlinux_get_current_pid $procid $sim $new_task_struct]
                    set ppid 0

                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "****************************** kthread_create  PID:$pid (PPID:$ppid)    function:$name "
                    }
                    set clone_flags 0
                    $sim processtree newproc $pid $ppid $clone_flags
                    $sim processtree exec $pid $name
                    EventLog::KernelExec $sim $procid $pid $name 
                } errs] } { puts "kthread_create errs are $errs" }

    return 0
}

#
#   This proc should be invoked on calls to setup_system 
#
#   This is a special case for labeling process 2
#
proc pidtracker_setup_system { sim args } {

    global vmlinux_structs
    global vmlinux_filename

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    if { [catch {
                    set name [vmlinux_get_current_name $procid $sim]
                    set pid [vmlinux_get_current_pid $procid $sim]

                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "****************************** setup_system  PID:$pid function:$name "
                    }
                    $sim processtree exec $pid $name
                    EventLog::KernelExec $sim $procid $pid $name
                } errs] } { puts "setup_system errs are $errs" }

    return 0
}

#
#   This proc should be on calls to kthreadd 
#
#   This is a special case for labeling process 2
#
proc pidtracker_kthreadd { sim args } {

    global vmlinux_structs
    global vmlinux_filename

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    if { [catch {
                    set pid [vmlinux_get_current_pid $procid $sim]

                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "****************************** kthreadd  PID:$pid function:kthreadd "
                    }

                    $sim processtree exec $pid "kthreadd"
                    EventLog::KernelExec $sim $procid $pid "kthreadd"
                } errs] } { puts "kthreadd errs are $errs" }

    return 0
}

#
#   Called when a kthread_stop is called 
#       r3 points to the task_struct
#
proc pidtracker_kthread_stop { sim args } {

    global vmlinux_structs
    global vmlinux_filename

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    if { [catch {
                    set task_struct [$sim cpu $procid display gpr 3]
                    set name [vmlinux_get_current_name $procid $sim $task_struct]
                    set pid [vmlinux_get_current_pid $procid $sim $task_struct]
                    set ppid 0

                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "****************************** kthread_stop  PID:$pid (PPID:$ppid)    function:$name "
                    }

                    $sim processtree killproc $pid 
                } errs] } { puts "kthread_stop errs are $errs" }

    return 0
}

proc load_struct_ptr { addr struct_entry_name {sim mysim} {procid 0:0:0} } {
    global vmlinux_structs

    set new_ea [expr $addr + $vmlinux_structs($struct_entry_name)]
    set new_ra [$sim cpu $procid dtranslate $new_ea ]
    set mem_val [$sim cpu $procid mem display $new_ra 8]

    return $mem_val

}

proc pidtracker_do_exit { sim args } {
    global vmlinux_structs

    # r3 = pointer to task_struct that is exiting

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    set code [format "%d" [$sim cpu $procid display gpr 3]]
    set pid [vmlinux_get_current_pid $procid $sim]
    set name [vmlinux_get_current_name $procid $sim]

    global process_track_verbose
    if { $process_track_verbose > 0 } {
        puts ".do_exit called code=$code for pid $pid name $name"
    }

    $sim processtree killproc $pid

    return 0
}

proc pidtracker_do_mmap_pgoff { sim args } {

    # Note: the pid/name offset is hardcoded and may changed
    #       with different kernel version. The offsets used
    #       in this code is derived from kernel version 2.6.7    
    global vmlinux_structs

    if { [catch "array set arginfo $args" errs] == 0 } {
        set pc $arginfo(address)
        set procid $arginfo(mcm):$arginfo(cpu):$arginfo(thread)
    } else {
        set pc 0
        set procid 0:0:0
    }

    if { [catch {
                    set pid [vmlinux_get_current_pid $procid $sim]

                    # gpr3 holds pointer to the filename to be exec'd
                    set file_ptr [format "%ld" [$sim cpu $procid display gpr 3]]
                    set addr [$sim cpu $procid display gpr 4]
                    set len [$sim cpu $procid display gpr 5]
                    set prot [$sim cpu $procid display gpr 6]
                    set flags [$sim cpu $procid display gpr 7]
                    set pgoff [$sim cpu $procid display gpr 8]

                    global process_track_verbose
                    if { $process_track_verbose > 0 } {
                        puts "do_mmap_pgoff called for file_ptr $file_ptr addr $addr len $len prot $prot flags $flags pgoff $pgoff"
                    }

                    if { $file_ptr != 0 } {
                        # find name

                        # Older Linux kernel's just have a f_dentry field in the file struct and newer ones have it 
                        # as part of the f_path struct.   Figure out which is which
                        if { [info exists vmlinux_structs(file.f_dentry)] } {
                            set dentry_ptr [load_struct_ptr $file_ptr file.f_dentry $sim $procid]
                        } else {
                            if { [info exists vmlinux_structs(file.f_path)] } {
                                set path_ptr [expr $file_ptr + $vmlinux_structs(file.f_path)]
                                set dentry_ptr [load_struct_ptr $path_ptr path.dentry $sim $procid]
                            } else {
                                # MDK: This message prints a gazillion times in run_cell_sdk11_sdk_tests,
                                # MDK: so I'm commenting it out until we have a better idea of what to do here.
                                #puts "TCL Script Error: Don't now how to read file path info for this kernel"
                                set dentry_ptr -1
                            }
                        }

                        if {$dentry_ptr >=0} {
                            # qstr structure is included in dentry struct
                            set name_ptr [expr $dentry_ptr + $vmlinux_structs(dentry.d_name)]
                            set name_ptr [load_struct_ptr $name_ptr qstr.name $sim $procid]
                            set name [get_string_at_ea $sim $name_ptr $procid]

                            $sim processtree addfile $pid $name $addr $len $prot $flags $pgoff
                        }
                    }
                } errs] } { 
                            puts "do_mmap_pgoff errs $errs" 
                          }

    return 0
}
