#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004, 2017
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : linux_track.tcl
##
## Purpose    : Mambo Linux process tracking
##
#####################################################
#####################################################

###################################################
# Package initialization
###################################################

namespace eval LinuxTrack {
    variable version            "1.0"
    variable vmlinux_structs
    variable vmlinux_filename   "vmlinux-not-loaded"
    variable task_struct
    variable debug              0

    array set ts_cache_suppress [list \
        "kthread" 1 \
        "khelper" 1 \
        "bash" 1 \
        "callthru" 1 \
        ]
}

package provide LinuxTrack $LinuxTrack::version
package require OSInfo

###################################################
# Process tracking triggers
###################################################

proc LinuxTrack::trigger_switch { sim args } {
    # We need to grab the args passed to __switch_to at the beginning
    # and then use them at the end so we can be more assured we find translations
    # for both task_struct's
#    puts "LinuxTrack::trigger_switch"

    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    # By convention, gpr3 holds pointer to task_struct of the old thread
    #set gpr3 [$sim cpu $mct display gpr 3]
    #set oldpid [vmlinux_get_current_pid $mct $sim $gpr3]
    #set oldname [vmlinux_get_current_name $mct $sim $gpr3]

    # By convention, gpr4 holds pointer to task_struct of the new thread    
#    puts "LinuxTrack::trigger_switch set tsp"
    set tsp [$sim cpu $mct display gpr 4]
    #puts "LinuxTrack::trigger_switch set pid"
    set pid [LinuxTrack::get_current_pid $mct $sim $tsp]
    #puts "LinuxTrack::trigger_switch set procname"
    set procname [LinuxTrack::get_current_name $mct $sim $tsp]
    #puts "LinuxTrack::trigger_switch OSInfo::Dispatch"
    OSInfo::Dispatch [list mct $mct pid $pid exec $procname]
    #puts "LinuxTrack::trigger_switch done"
}

proc LinuxTrack::trigger_execve { sim args } {
    # Note: the pid/name offset is hardcoded and may changed
    #       with different kernel version. The offsets used
    #       in this code is derived from kernel version 2.6.7    

    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    # gpr3 holds pointer to the filename to be exec'd
    set ea [$sim cpu $mct display gpr 3]
    set ra [$sim cpu $mct dtranslate $ea]
    set filename [$sim cpu $mct memory display $ra STRING]     

    OSInfo::Dispatch [list mct $mct exec $filename]
}

proc LinuxTrack::trigger_exit { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    # r3 = exit code
    set code [format "%d" [$sim cpu $mct display gpr 3]]
    set pid [LinuxTrack::get_current_pid $mct $sim]
    set procname [LinuxTrack::get_current_name $mct $sim]
}

proc LinuxTrack::trigger_copyproc { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    # If on return it is not an error (ie gpr3 u< -4095) then get the results
    set tsp [$sim cpu $mct display gpr 3]
    if { $tsp < 0xFFFFFFFFFFFFF000 } {
        #set ppid [LinuxTrack::get_current_pid $mct $sim]
        set pid [LinuxTrack::get_current_pid $mct $sim $tsp]
        OSInfo::Dispatch [list mct $mct pid $pid]
    }
}

proc LinuxTrack::trigger_ktcreate { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    set tsp [$sim cpu $mct display gpr 3]
    set procname [LinuxTrack::get_current_name $mct $sim $tsp]
    set pid [LinuxTrack::get_current_pid $mct $sim $tsp]
    #set ppid [LinuxTrack::get_current_ppid $mct $sim $tsp]
    OSInfo::Dispatch [list mct $mct pid $pid exec $procname]
}

proc LinuxTrack::trigger_kthreadd { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    set pid [LinuxTrack::get_current_pid $mct $sim]
    OSInfo::Dispatch [list mct $mct pid $pid exec "kthreadd"]
}

proc LinuxTrack::trigger_setupsys { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    set pid [LinuxTrack::get_current_pid $mct $sim]
    set procname [LinuxTrack::get_current_name $mct $sim]
    OSInfo::Dispatch [list mct $mct pid $pid exec $procname]
}

proc LinuxTrack::trigger_ktstop { sim args } {
    if { [catch "array set p $args" errs] == 0 } {
        set pc $p(address)
        set mct $p(mcm):$p(cpu):$p(thread)
    } else {
        set pc 0
        set mct 0:0:0
    }

    set tsp [$sim cpu $mct display gpr 3]
    set procname [LinuxTrack::get_current_name $mct $sim $tsp]
    set pid [LinuxTrack::get_current_pid $mct $sim $tsp]
    #set ppid [LinuxTrack::get_current_ppid $mct $sim $tsp]
}


###################################################
# Process tracking trigger setters
###################################################

#
# Set a trigger at the START of a specified vmlinux function
#
proc LinuxTrack::set_trigger { function_name { tcl_proc "break_point mysim $function_name" } { sim mysim } } {
    variable vmlinux_filename 
    #puts "LinuxTrack::set_trigger"
    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
	#puts "LinuxTrack::set_trigger catch 1"
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
	    #puts "LinuxTrack::set_trigger catch 2"
            if { [info exists func_info(FUNC_START)] } {
		#puts "LinuxTrack::set_trigger catch 3"
                set entry_addr $func_info(FUNC_START)
		#puts "LinuxTrack::set_trigger set entry_addr"
                #puts "Calling LinuxTrack::set_trigger $sim trigger set pc $entry_addr $tcl_proc"


                $sim trigger set pc $entry_addr $tcl_proc
		#puts "Called LinuxTrack::set_trigger $sim trigger set pc $entry_addr $tcl_proc"
            }
        } else {
            return -code error "LinuxTrack::set_trigger failed with $errs"
        }
    } else {
        return -code error "LinuxTrack::set_trigger object call failed with $errs"
    }
}

#
# Set a trigger on all RETURNS (blr) from a specified vmlinux function
#
proc LinuxTrack::set_trigger_on_return { function_name { tcl_proc "break_point mysim $function_name" } { sim mysim } } {
    variable vmlinux_filename 

    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
            if { [info exists func_info(FUNC_START)] } {
                set inst_data [object inst_data $vmlinux_filename $func_info(FUNC_START) $func_info(FUNC_END)]
                array set ea_inst $inst_data
                foreach ea [array names ea_inst] {
                    if { $ea_inst($ea) == 0x4e800020 } {
                        $sim trigger set pc $ea $tcl_proc
                    }
                }
            }
        } else {
            return -code error "LinuxTrack::set_trigger_on_return failed with $errs"
        }
    } else {
        return -code error "LinuxTrack::set_trigger_on_return object call failed with $errs"
    }
}

###################################################
# Map file manipulation
###################################################

proc LinuxTrack::get_struct_info { filename } {
    variable vmlinux_structs

    #
    #   Execute the parse_dwarf.pl command to return the tcl file that contains the structure info
    #
    global env
    set parse_dwarf_pgm $env(EXEC_DIR)/parse_dwarf.pl

    if { [catch {
            set tclfile [exec $parse_dwarf_pgm $filename]
            source $tclfile
        } errs ] } { 
        return -code error "LinuxTrack::get_struct_info got errors: $errs" 
    }
}

proc LinuxTrack::ts_lookup { field type size { mct 0:0:0 } { sim mysim } { tsp empty } } {
    variable vmlinux_structs
    variable task_struct
    variable debug
    variable ts_cache_suppress

    if {$debug} {
        puts "LinuxTrack::ts_lookup(field=$field, type=$type, size=$size, mct=$mct, sim=$sim, tsp=$tsp)"
    }

    # If the TSP is "null", we go get the current one
    if { $tsp eq "empty" } {

        # Older kernels had a paca_struct.xCurrent field and newer ones have paca_struct.__current field
        if { [info exists vmlinux_structs(paca_struct.__current)] } {
            set member "__current"
        } elseif { [info exists vmlinux_structs(paca_struct.xCurrent)] } {
            set member "xCurrent"
        } else {
            return 0
        }

        set base [$sim cpu $mct display gpr 13]
        set ea [format "0x%016lX" [expr $base + $vmlinux_structs(paca_struct.$member)]]
        set ra [$sim cpu $mct dtranslate $ea ]
        set tsp [$sim cpu $mct mem display $ra 8]

        if {$debug} {
            puts "LinuxTrack::ts_lookup -- empty tsp base=$base, ea=$ea, ra=$ra, tsp=$tsp"
        }
    }

    # If we already have a value for this field in the cached task struct...
    # DBM: Disabled the cache, since the same tsp seems to be shared
    if { 0 && [info exists task_struct($sim.$tsp.$field)] } {
        if {$debug} {
            puts "LinuxTrack::ts_lookup -- found task_struct($sim.$tsp.$field)=$task_struct($sim.$tsp.$field)"
        }
        return $task_struct($sim.$tsp.$field)
    }

    # If we don't have any information about this field in the file...
    if { ! [info exists vmlinux_structs(task_struct.$field)] } {
        return -code error "LinuxTrack::ts_lookup(field=$field type=$type size=$size mct=$mct sim=$sim tsp=$tsp): unknown task_struct field '$field'" 
        return 0
    }

    # Otherwise, we translate the EA of the field...
    if { [catch { 
            set ea [format "0x%016lX" [expr $tsp + $vmlinux_structs(task_struct.$field)]]
            set ra [$sim cpu $mct dtranslate $ea]
        } errs] } { 
            return -code error "LinuxTrack::ts_lookup(field=$field type=$type size=$size mct=$mct sim=$sim tsp=$tsp): could not translate ea=$ea ($errs)" 
        }

    # And look up the value...
    set rc 0
    #set suppress_cache 0
    switch -exact -- $type {
        int {
            if { [catch { 
                    set val [format "%ld" [$sim cpu $mct mem display $ra $size]]
                } errs] } { 
                    set rc 1
                }
        }
        ptr {
            if { [catch { 
                    set val [format "0x%016lX" [$sim cpu $mct mem display $ra 8]]
                } errs] } { 
                    set rc 1
                }
        }
        string {
            if { [catch { 
                    set val [$sim cpu $mct memory display $ra STRING]     
                    #set suppress_cache [info exists ts_cache_suppress($val)]
                } errs] } { 
                    set rc 1
                }
        }
        default {
            set rc 1
            set errs "unknown type='$type'"
        }
    }

    if {$rc} {
        return -code error "LinuxTrack::ts_lookup(field=$field type=$type size=$size mct=$mct sim=$sim tsp=$tsp): $errs" 
    } else {
        if {$debug} {
            puts "LinuxTrack::ts_lookup -- results ea=$ea, ra=$ra, val=$val to task_struct($sim.$tsp.$field)"
        }
        # DBM: Disabled the cache...
        # set task_struct($sim.$tsp.$field) $val
        return $val
    }
}

proc LinuxTrack::get_current_pid { { mct 0:0:0 } { sim mysim } { tsp empty } } {
    return [LinuxTrack::ts_lookup pid int 4 $mct $sim $tsp]
}

proc LinuxTrack::get_current_ppid { { mct 0:0:0 } { sim mysim } { tsp empty } } {
    set tsp [LinuxTrack::ts_lookup parent ptr 8 $mct $sim $tsp]
    return [LinuxTrack::get_current_pid $mct $sim $tsp]
}

proc LinuxTrack::get_current_name { { mct 0:0:0 } { sim mysim } { tsp empty } } {
    return [LinuxTrack::ts_lookup comm string 0 $mct $sim $tsp]
}


###################################################
# Process tracking
###################################################

proc LinuxTrack::enable {sim filename {address 0} {size 0}} {
    variable vmlinux_filename [file tail $filename]
    variable vmlinux_dirname [file dirname $filename]
    object dir $vmlinux_dirname
#    variable vmlinux_filename $filename
    #puts "Calling LinuxTrack::get_struct_info $filename"
    LinuxTrack::get_struct_info $filename
    #puts "Called LinuxTrack::get_struct_info $filename"
    # Setup up the PID tracking

    LinuxTrack::set_trigger .__switch_to [list LinuxTrack::trigger_switch $sim]
    #puts "LinuxTrack::set_trigger .do_execve"
    LinuxTrack::set_trigger .do_execve [list LinuxTrack::trigger_execve $sim]
    #puts "LinuxTrack::set_trigger .compat_do_execve"
    LinuxTrack::set_trigger .compat_do_execve [list LinuxTrack::trigger_execve $sim]
    #LinuxTrack::set_trigger .do_exit [list LinuxTrack::trigger_exit $sim]
    #puts "LinuxTrack::set_trigger_on_return .copy_process"
    LinuxTrack::set_trigger_on_return .copy_process [list LinuxTrack::trigger_copyproc $sim]
    #puts "LinuxTrack::set_trigger_on_return .kthread_create"
    LinuxTrack::set_trigger_on_return .kthread_create [list LinuxTrack::trigger_ktcreate $sim]
    #puts "LinuxTrack::set_trigger .kthreadd"
    LinuxTrack::set_trigger .kthreadd [list LinuxTrack::trigger_kthreadd $sim]
    #puts "LinuxTrack::set_trigger .setup_system"
    LinuxTrack::set_trigger .setup_system [list LinuxTrack::trigger_setupsys $sim]
    #LinuxTrack::set_trigger .kthread_stop [list LinuxTrack::trigger_ktstop $sim]
}
