#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004, 2017
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : Step.tcl
##
## Purpose    : Enhanced Mambo step functions
##
#####################################################
#####################################################

namespace eval Step {
    variable version            "1.1"
    variable user_mode_icount   0
    variable priv_mode_icount   0
    variable prev_icount        0
    variable user_mode_cycles   0
    variable priv_mode_cycles   0
    variable prev_cycle         0
    variable mode               0
    variable cpu                0
    variable thread             0
    variable mct                "0:0:0"
    variable trig_id
    variable dispatch_state
    variable dispatch_focus
    variable dispatch_callback_id
    variable dispatch_interrupt
    variable debug              0
    variable silent             0
}

package provide Step $Step::version
package require BCast

####
# Return the current mode, given an msr value
#
proc Step::get_mode { msr } {
    if {[expr $msr & 0x4000]} {
        # User mode
        return 1
    } else {
        # Supervisor mode
        return 0
    }
}

####
# Set the msr change trigger
#
proc Step::set_msr_trigger { sim cpu thread args } {
    variable trig_id
    variable debug

    if {[llength [array names trig_id $cpu,$thread]] == 0} {
        set trig_id($cpu,$thread) [eval $sim cpu $cpu thread $thread trigger spr msr change $args]
    } else {
        eval $sim cpu $cpu thread $thread trigger spr msr id $trig_id($cpu,$thread) change $args
    }

    if {$debug} {
        puts "Step::set_msr_trigger [$sim cpu $cpu thread $thread trigger spr msr]"
    }
}

####
# Clear the msr change trigger
#
proc Step::clear_msr_trigger { sim cpu thread args } {
    variable trig_id
    variable debug

    if {[llength [array names trig_id $cpu,$thread]] != 0} {
        if {$debug} {
            puts "Step::clear_msr_trigger [$sim cpu $cpu thread $thread trigger spr msr]"
        }
        eval $sim cpu $cpu thread $thread trigger spr msr id $trig_id($cpu,$thread) clear
        array unset trig_id $cpu,$thread
    }

}

####
# Reset all icount state variables
#
proc Step::icount_reset { sim pmct } {
    variable cpu
    variable thread

    if {![regexp {((\d):)?(\d):(\d)} $pmct match x m c t]} {
        return -code error "Step::icount_reset -- malformed mcm:cpu:thread id '$pmct'"
    }

    variable mct $pmct
    variable dispatch_state
    if {![info exists dispatch_state($mct)]} {
        set dispatch_state($mct) 1
    }
    variable dispatch_interrupt
    if {![info exists dispatch_interrupt($mct)]} {
        set dispatch_interrupt($mct) 0
    }

    set cur_icount [$sim cpu $cpu thread $thread display instruction_count]
    set msr    [$sim cpu $cpu thread $thread display spr msr]

    variable user_mode_icount 0
    variable priv_mode_icount 0
    variable prev_icount $cur_icount
    variable user_mode_cycles 0
    variable priv_mode_cycles 0
    variable prev_cycle [$sim display cycle]
    variable mode [Step::get_mode $msr]
}

####
# Determine the privilege and user mode instruction
# counts made since the last call to this function.
# Accumulate total counts in user_mode_icount and
# priv_mode_icount
#
proc Step::icount_tally { sim msr } {
    variable cpu
    variable thread
    variable mode
    variable user_mode_icount
    variable priv_mode_icount
    variable prev_icount
    variable user_mode_cycles
    variable priv_mode_cycles
    variable prev_cycle

    set to_mode [Step::get_mode $msr]

    set cur_icount [$sim cpu $cpu thread $thread display instruction_count]
    set cur_cycle [$sim display cycle]

    if {$mode} {
        # Accumulate user mode icount and cycles
        set user_mode_icount [format "%ld" [expr $user_mode_icount+($cur_icount-$prev_icount)]]
        set user_mode_cycles [format "%ld" [expr $user_mode_cycles+($cur_cycle-$prev_cycle)]]
    } else {
        # Accumulate supervisor mode icount and cycles
        set priv_mode_icount [format "%ld" [expr $priv_mode_icount+($cur_icount-$prev_icount)]]
        set priv_mode_cycles [format "%ld" [expr $priv_mode_cycles+($cur_cycle-$prev_cycle)]]
    }

    set prev_icount $cur_icount
    set prev_cycle $cur_cycle

    set mode $to_mode
}

####
# Display the current user and privilege mode icount
#
proc Step::icount_display { {sim mysim} } {
    variable cpu
    variable thread
    variable user_mode_icount
    variable priv_mode_icount
    variable user_mode_cycles
    variable priv_mode_cycles
    variable silent

    Step::icount_tally $sim [$sim cpu $cpu thread $thread display spr msr]
    if {!$silent} {
        puts [format "*** ICOUNT: User=%d, Priv=%d" $user_mode_icount $priv_mode_icount]
        #puts [format "*** CYCLES: User=%d, Priv=%d" $user_mode_cycles $priv_mode_cycles]
    }
}

####
# Callback function for msr register change trigger. This
# will in turn call $func
#
proc Step::trigger_proxy { func args } {
    variable dispatch_state
    variable mct

    if {$dispatch_state($mct)} {
        array set p $args
        if {[catch {eval $func $p(val)} result]} {
            puts "Step::trigger_proxy: '$func $p(val)' could not be evaluated"
            puts "Step::trigger_proxy: result was '$result'"
            exit
        }
    }
}

####
# Simstop if the msr privilege mode matches the given value
#
proc Step::stop_on_mode { sim mode msr } {
    variable debug

    if {[Step::get_mode $msr] == $mode} {
        if {$debug} {
            puts "Step::stop_on_mode: Reached mode $mode"
            #simdebug set instruction $mode
        }
        simstop
    }
}

####
# Run until we are in the given mode
#
proc Step::run_until_mode { sim mode } {
    variable cpu
    variable thread
    variable dispatch_state
    variable dispatch_interrupt
    variable mct
    variable debug

    # If not already in desired mode...
    set curr_mode [Step::get_mode [$sim cpu $cpu thread $thread display spr msr]]
    if {$curr_mode != $mode} {
        # Set up to interrupt our "go" when we change modes or dispatch state
        Step::set_msr_trigger $sim $cpu $thread [list Step::trigger_proxy [list Step::stop_on_mode $sim $mode]] 0x4000
        # The dispatch interrupt should be needed only for kernel processs tracking. For user process
        # tracking, we are already being interrupted for kernel mode switches and therefore should not miss
        # a dispatch state change (kernel is needed to switch processes)
        if {$mode == 0} {
            set dispatch_interrupt($mct) 1
        }

        if {$debug} {
            puts "Step::run_until_mode: Advancing cpu $cpu thread $thread to mode $mode..."
        }

        while {($dispatch_state($mct) == 0) || ($curr_mode != $mode)} {
            $sim go
            set curr_mode [Step::get_mode [$sim cpu $cpu thread $thread display spr msr]]
            if {$debug} {
                puts "Step::run_until_mode: $sim go interrupted, dispatch $dispatch_state($mct) desired mode $mode, current mode $curr_mode"
            }
        }
        set dispatch_interrupt($mct) 0

        Step::clear_msr_trigger $sim $cpu $thread

        Step::icount_tally $sim [$sim cpu $cpu thread $thread display spr msr]

        if {$debug} {
            variable user_mode_icount
            variable priv_mode_icount
            puts "Step::run_until_mode: Mode $mode reached, ucount=$user_mode_icount, kcount=$priv_mode_icount"
            #simdebug set instruction $mode
        }
    } elseif {$debug} {
        puts "Step::run_until_mode: Already in mode $mode..."
    }
}

####
# User-mode instruction step
#
proc Step::ustep { count {mct "0:0:0"} {sim mysim} } {
    variable cpu
    variable thread
    variable user_mode_icount
    variable debug
    variable silent

    Step::icount_reset $sim $mct

    while { $user_mode_icount < $count } {
        Step::run_until_mode $sim 1

        if { $user_mode_icount < $count } {
            if {$debug} {
                puts "Step::ustep -- stepping [format %ld [expr $count-$user_mode_icount]]"
            }

            Step::set_msr_trigger $sim $cpu $thread [list Step::trigger_proxy [list Step::stop_on_mode $sim 0]] 0x4000
            $sim cpu $cpu thread $thread step [expr $count-$user_mode_icount]
            Step::clear_msr_trigger $sim $cpu $thread

            Step::icount_tally $sim [$sim cpu $cpu thread $thread display spr msr]

            if {$debug} {
                puts "Step::ustep -- done stepping ucount=$user_mode_icount"
            }
        }
    }

    if {!$silent && [display regress_mode] == "on"} {
        # In regress mode, only display the user mode icount
        puts [format "*** ICOUNT: User=%d" $user_mode_icount]
    } else {
        Step::icount_display $sim
    }
}

####
# Privilege-mode instruction step
#
proc Step::kstep { count {mct "0:0:0"} {sim mysim} } {
    variable cpu
    variable thread
    variable priv_mode_icount
    variable debug
    variable silent

    Step::icount_reset $sim $mct

    while { $priv_mode_icount < $count } {
        Step::run_until_mode $sim 0

        if { $priv_mode_icount < $count } {
            if {$debug} {
                puts "Step::kstep -- stepping [format %ld [expr $count-$priv_mode_icount]]"
            }

            Step::set_msr_trigger $sim $cpu $thread [list Step::trigger_proxy [list Step::stop_on_mode $sim 1]] 0x4000
            $sim cpu $cpu thread $thread step [expr $count-$priv_mode_icount]
            Step::clear_msr_trigger $sim $cpu $thread

            Step::icount_tally $sim [$sim cpu $cpu thread $thread display spr msr]

            if {$debug} {
                puts "Step::kstep -- done stepping kcount=$priv_mode_icount"
            }
        }
    }

    if {!$silent && [display regress_mode] == "on"} {
        # In regress mode, only display the priv mode icount
        puts [format "*** ICOUNT: Priv=%d" $priv_mode_icount]
    } else {
        Step::icount_display $sim
    }
}

####
# Dispatch -- intercept dispatch calls from the process tracker
#
proc Step::dispatch { args } {
    variable dispatch_state
    variable dispatch_focus
    variable debug

    array set p $args
    set sim $p(sim)
    set mct $p(mct)
    set exec $p(exec)

    variable prev_icount
    variable prev_cycle

    set cur_icount [$sim cpu $mct display instruction_count]
    set cur_cycle [$sim display cycle]

    if {$debug} {
        puts "Step::dispatch mct $mct exec $exec icount $cur_icount"
    }

    set old_state $dispatch_state($mct)
    if {[string match $dispatch_focus($mct) $exec]} {
        set prev_icount $cur_icount
        set prev_cycle  $cur_cycle
        set dispatch_state($mct) 1
        if {$debug} {
            puts "Step::dispatch MATCH focal $exec"
        }
    } elseif {$dispatch_state($mct)} {
        Step::icount_tally $sim [$sim cpu $mct display spr msr]
        set dispatch_state($mct) 0
    }

    variable dispatch_interrupt
    if {$dispatch_interrupt($mct) && ($old_state != $dispatch_state($mct))} {
        simstop
    }

}

####
# Focus -- set which process name to restrict stepping to
#
proc Step::focus { sim mct {f none} } {
    variable dispatch_state
    variable dispatch_focus
    variable dispatch_callback_id
    variable dispatch_interrupt

    switch -exact -- $f {
        none {
            if {[info exists dispatch_callback_id($mct)]} {
                BCast::Ignore $dispatch_callback_id($mct) exec
            }
            set dispatch_state($mct) 1
            set dispatch_focus($mct) "none"
            puts "Step::focus -- \[$mct\] Remove step count process restrictions"
        }
        default {
            if {![info exists dispatch_callback_id($mct)]} {
                set dispatch_callback_id($mct) [BCast::Register $mct [list Step::dispatch sim $sim]]
            }
            BCast::Listen exec $dispatch_callback_id($mct)
            set dispatch_state($mct) 0
            set dispatch_focus($mct) $f
            puts "Step::focus -- \[$mct\] Restrict step counts to process $f"
        }
    }

    set dispatch_interrupt($mct) 0
}
